/*----------------------------------------------------------
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version
2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied
warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public
License along with this program; if not, write to the Free
Software Foundation, Inc., 59 Temple Place, Suite 330,
Boston, MA  02111-1307  USA

The original author of this program is Hin Jang
March 20, 2003
-----------------------------------------------------------*/

// ----------------------------------------------------------
// catmullClark.cpp (for quad meshes only)
// ----------------------------------------------------------

#include "catmullClark.h"

CatmullClark::CatmullClark(KDTreeRoot *k)
{

  forest = k;

}

CatmullClark::~CatmullClark()
{

  delete forest;
  adjacentEdgePoints.clear();

}

// returns non-zero on success
//

int CatmullClark::load(char *fileName)
{

  int numTokens;

  string line;
  vector<string> s;

  Parser *par;
  Polygon *p;
  Vertex *v;

  ifstream file(fileName);
  if (file.is_open() == false) return 0;

  par = new Parser;
  numTokens = 0;
  while (getline(file, line, '\n')) {

    numTokens = par->split(line, " ");
    if (numTokens > 0) {

      s = par->getVector();
      switch(line[0]) {
	
      case 'v':

	if (s.size() != 13) {

	  cerr << "Invalid vertex definition in "
	       << fileName << endl;

	  file.close();
	  delete par;
	  return 0;

	}

	v = new Vertex();

	// Cartesian coordinates
	//

	v->p[0] = strtod(s[1].c_str(), NULL);
	v->p[1] = strtod(s[2].c_str(), NULL);
	v->p[2] = strtod(s[3].c_str(), NULL);

	// RGBA colour
	//

	v->c[0] = strtod(s[4].c_str(), NULL);
	v->c[1] = strtod(s[5].c_str(), NULL);
	v->c[2] = strtod(s[6].c_str(), NULL);
	v->c[3] = strtod(s[7].c_str(), NULL);

	// normal
	//

	v->n[0] = strtod(s[8].c_str(), NULL);
	v->n[1] = strtod(s[9].c_str(), NULL);
	v->n[2] = strtod(s[10].c_str(), NULL);

	// unique identifiers
	//

	v->id = atoi(s[11].c_str());
	v->tag = atoi(s[12].c_str());

	if (v->tag < Vertex::Regular ||
	    v->tag > Vertex::ConcaveCorner) {

	  v->tag = Vertex::Regular;

	}

	v->subLevel = 0;
	v->pid = v->id;   

	forest->augmentVertexPool(v);
	break;
	
      case 'p':

	if (s.size() != 10) {

	  cerr << "Invalid polygon definition in "
	       << fileName << endl;

	  file.close();
	  delete par;
	  return 0;

	}

	p = new Polygon();

	// vertices
	//

	p->v[0] = atoi(s[1].c_str());
	p->v[1] = atoi(s[2].c_str());
	p->v[2] = atoi(s[3].c_str());
	p->v[3] = atoi(s[4].c_str());

	// neighbours
	//

	p->n[0] = atoi(s[5].c_str());
	p->n[1] = atoi(s[6].c_str());
	p->n[2] = atoi(s[7].c_str());
	p->n[3] = atoi(s[8].c_str());

	// unique identifiers
	//

	p->id  = atoi(s[9].c_str());
	p->pid = p->id;

	forest->augmentPolygonPool(p);
	break;
	
      default:
	break;
	
	
      }

    }

  }

  file.close();
  delete par;

  forest->create();

  vertexPool  = forest->getVertexPool();
  polygonPool = forest->getPolygonPool();

  return 1;

}

void CatmullClark::draw(int subLevel)
{

  if (subLevel < 0) return;
  if (subLevel > forest->getMaxSubLevel()) {

    applySubdivisionRules(subLevel);

  }

  forest->traverse(subLevel);

}

void
CatmullClark::applySubdivisionRules(int s)
{

  Vertex *u;
  Polygon *p, *g;
  vector<int> polyList;
  vector<int> vertList;
  vector<int> newFacePoints;
  vector<int> newEdgePoints;
  int newVertexID;

  int i, j, k, m, pmax, vmax, fmax;

  if (forest->getMaxSubLevel() < s) {

    forest->setMaxSubLevel(s);

  }

  pmax = polygonPool.size();
  vmax = vertexPool.size();

  adjacentEdgePoints.clear();
  for (i=0; i<vmax; ++i) {

    // scan through polygonPool to determine which
    // edges and faces are incident/adjacent to
    // vertex 'i'
    //

    vertList.clear();
    polyList.clear();
    for (j=0; j<pmax; ++j) {

      p = polygonPool[j];

      for (k=0; k<4; ++k) {

	u = vertexPool[p->v[k]];

	if (p->v[k] == i &&
	    u->subLevel == s-1) {

	  vertList.push_back(p->v[(k+1)&3]);
	  polyList.push_back(p->id);
	  break;

	}

      }

    }

    if (vertList.size() > 1 && polyList.size() > 1) {

      // apply all masks to calculate surface points
      //

      newFacePoints=faceMask(polyList, s);
      newEdgePoints=edgeMask(vertList, newFacePoints,
			     polyList, i, s);
      newVertexID=vertexMask(vertList, newFacePoints, i, s);

      // add polygon(s) to children
      //

      fmax = newFacePoints.size();
      for (m=0; m<fmax; ++m) {

	g = new Polygon();
	g->v[0] = newVertexID;
	g->v[1] = newEdgePoints[m];
	g->v[2] = newFacePoints[m];
	g->v[3] = adjacentEdgePoints[g->v[2]];
	g->id   = polygonPool.size()+1;
	g->pid  = vertexPool[g->v[2]]->pid;

	polygonPool.push_back(g);
	forest->add(s, g);

      }


    }

  }

  forest->setVertexPool(vertexPool);
  forest->setPolygonPool(polygonPool);

}

// returns integer greater than or equal to zero
// on success
//

int CatmullClark::findDuplicateVertex(Vertex *v)
{

  Vertex *u;
  int i, n;

  n = vertexPool.size();

  for (i=0; i<n; ++i) {

    u = vertexPool[i];
    if (u->p[0] == v->p[0] &&
	u->p[1] == v->p[1] &&
	u->p[2] == v->p[2]) {

      return u->id;

    }

  }

  return -1;

}

int CatmullClark::findAdjacentFaceID(vector<int> polyList,
				     int e0, int e1)
{

  Polygon *p;
  int i, j, n, prev;

  // scan through polyList to locate the
  // face point adjacent to edge (e0, e1)
  //

  n = polyList.size();
  for (i=0; i<n; ++i) {

    p = polygonPool[polyList[i]-1];

    prev = 3;
    for (j=0; j<4; ++j) {

      if (p->v[j] == e0 && p->v[prev] == e1) {

	return i;
	
      }
      
      prev = j;

    }
    
  }

  return n-1;

}


vector<int> CatmullClark::faceMask(vector<int> polyList,
				   int subLevel)
{

  vector<int> newVertices;
  Vertex *v, *nv;
  Polygon *p;

  int i, j, n, dupID;
  
  newVertices.clear();
  n = polyList.size();


  for (i=0; i<n; ++i) {

    nv = new Vertex();
    p = polygonPool[polyList[i]-1];

    for (j=0; j<4; ++j) {

      v = vertexPool[p->v[j]];

      nv->p[0] += v->p[0];
      nv->p[1] += v->p[1];
      nv->p[2] += v->p[2];

      nv->c[0] += v->c[0];
      nv->c[1] += v->c[1];
      nv->c[2] += v->c[2];
      nv->c[3] += v->c[3];

      nv->n[0] += v->n[0];
      nv->n[1] += v->n[1];
      nv->n[2] += v->n[2];

    }

    nv->p[0] *= 0.25;
    nv->p[1] *= 0.25;
    nv->p[2] *= 0.25;
    
    nv->c[0] *= 0.25;
    nv->c[1] *= 0.25;
    nv->c[2] *= 0.25;
    nv->c[3] *= 0.25;
    
    nv->n[0] *= 0.25;
    nv->n[1] *= 0.25;
    nv->n[2] *= 0.25;

    dupID = findDuplicateVertex(nv);
    if (dupID < 0) {

      nv->id = vertexPool.size();
      nv->tag = 0;
      nv->subLevel = subLevel;
      nv->pid = p->pid;
      
      newVertices.push_back(nv->id);
      vertexPool.push_back(nv);

    } else {

      newVertices.push_back(dupID);

    }


  }

  return newVertices;

}

vector<int> CatmullClark::edgeMask(vector<int> vertList,
				   vector<int> newFacePoints,
				   vector<int> polyList,
				   int vertexID,
				   int subLevel)
{

  vector<int> newVertices;
  vector<int> faceList;

  Vertex *v, *nv, *e, *f0, *f1;
  int i, n, id, dupID;
  
  newVertices.clear();
  v = vertexPool[vertexID];

  n = newFacePoints.size();
  for (i=0; i<n; ++i) {
      
    e  = vertexPool[vertList[i]];
    nv = new Vertex();

    f0 = vertexPool[newFacePoints[i]];
    id = findAdjacentFaceID(polyList, vertexID, e->id);
    f1 = vertexPool[newFacePoints[id]];

    if (v->tag == Vertex::Crease &&
	e->tag == Vertex::Crease) {

      nv->p[0] = (v->p[0] + e->p[0]) * 0.5;
      nv->p[1] = (v->p[1] + e->p[1]) * 0.5;
      nv->p[2] = (v->p[2] + e->p[2]) * 0.5;
      
      nv->c[0] = (v->c[0] + e->c[0]) * 0.5;
      nv->c[1] = (v->c[1] + e->c[1]) * 0.5;
      nv->c[2] = (v->c[2] + e->c[2]) * 0.5;
      nv->c[3] = (v->c[3] + e->c[3]) * 0.5;
      
      nv->n[0] = (v->n[0] + e->n[0]) * 0.5;
      nv->n[1] = (v->n[1] + e->n[1]) * 0.5;
      nv->n[2] = (v->n[2] + e->n[2]) * 0.5;
      nv->tag = Vertex::Crease;

    } else {
      
      nv->p[0] = (v->p[0]+e->p[0]+f0->p[0]+f1->p[0]) * 0.25;
      nv->p[1] = (v->p[1]+e->p[1]+f0->p[1]+f1->p[1]) * 0.25;
      nv->p[2] = (v->p[2]+e->p[2]+f0->p[2]+f1->p[2]) * 0.25;
      
      nv->c[0] = (v->c[0]+e->c[0]+f0->c[0]+f1->c[0]) * 0.25;
      nv->c[1] = (v->c[1]+e->c[1]+f0->c[1]+f1->c[1]) * 0.25;
      nv->c[2] = (v->c[2]+e->c[2]+f0->c[2]+f1->c[2]) * 0.25;
      nv->c[3] = (v->c[3]+e->c[3]+f0->c[3]+f1->c[3]) * 0.25;
      
      nv->n[0] = (v->n[0]+e->n[0]+f0->n[0]+f1->n[0]) * 0.25;
      nv->n[1] = (v->n[1]+e->n[1]+f0->n[1]+f1->n[1]) * 0.25;
      nv->n[2] = (v->n[2]+e->n[2]+f0->n[2]+f1->n[2]) * 0.25;
      nv->tag = Vertex::Regular;
      
    }

    dupID = findDuplicateVertex(nv);
    if (dupID < 0) {

      nv->id = vertexPool.size();
      nv->subLevel = subLevel;
      nv->pid = f0->pid;

      newVertices.push_back(nv->id);
      vertexPool.push_back(nv);
      adjacentEdgePoints[f1->id] = nv->id;

    } else {

      newVertices.push_back(dupID);
      adjacentEdgePoints[f1->id] = dupID;

    }

    
  }
  
  return newVertices;

}

int CatmullClark::vertexMask(vector<int> vertList,
			     vector<int> newFacePoints,
			     int vertexID,
			     int subLevel)
{


  int newVertexID;
  Vertex *u, *v, *nv, *nv0, *nv1;
  int i, n, k, dupID;
  double d;

  v   = vertexPool[vertexID];
  nv  = new Vertex();
  nv0 = new Vertex();
  nv1 = new Vertex();

  n = vertList.size();

  if (v->tag == Vertex::Crease) {

    k = 0;
    for (i=0; i<n && k<2; ++i) {

      u = vertexPool[vertList[i]];
      if (u->tag == Vertex::Crease) {

	nv->p[0] += u->p[0] * 0.125;
	nv->p[1] += u->p[1] * 0.125;
	nv->p[2] += u->p[2] * 0.125;

	nv->c[0] += u->c[0] * 0.125;
	nv->c[1] += u->c[1] * 0.125;
	nv->c[2] += u->c[2] * 0.125;
	nv->c[3] += u->c[3] * 0.125;

	nv->n[0] += u->n[0] * 0.125;
	nv->n[1] += u->n[1] * 0.125;
	nv->n[2] += u->n[2] * 0.125;
	++k;

      }

    }

    nv->p[0] += v->p[0] * 0.75;
    nv->p[1] += v->p[1] * 0.75;
    nv->p[2] += v->p[2] * 0.75;

    nv->c[0] += v->c[0] * 0.75;
    nv->c[1] += v->c[1] * 0.75;
    nv->c[2] += v->c[2] * 0.75;
    nv->c[3] += v->c[3] * 0.75;
    
    nv->n[0] += v->n[0] * 0.75;
    nv->n[1] += v->n[1] * 0.75;
    nv->n[2] += v->n[2] * 0.75;
    nv->tag = Vertex::Crease;

  } else {

    // compute (average of vertList) / n
    //

    for (i=0; i<n; ++i) {

      v = vertexPool[vertList[i]];
      nv0->p[0] += v->p[0];
      nv0->p[1] += v->p[1];
      nv0->p[2] += v->p[2];

      nv0->c[0] += v->c[0];
      nv0->c[1] += v->c[1];
      nv0->c[2] += v->c[2];
      nv0->c[3] += v->c[3];

      nv0->n[0] += v->n[0];
      nv0->n[1] += v->n[1];
      nv0->n[2] += v->n[2];
      
    }

    d = 1.0 / (double)(n*n);
    nv0->p[0] *= d;
    nv0->p[1] *= d;
    nv0->p[2] *= d;
    
    nv0->c[0] *= d;
    nv0->c[1] *= d;
    nv0->c[2] *= d;
    nv0->c[3] *= d;
    
    nv0->n[0] *= d;
    nv0->n[1] *= d;
    nv0->n[2] *= d;
    
    n = newFacePoints.size();

    // compute (average of newFacePoints) / n
    //

    for (i=0; i<n; ++i) {
      
      v = vertexPool[newFacePoints[i]];
      
      nv1->p[0] += v->p[0];
      nv1->p[1] += v->p[1];
      nv1->p[2] += v->p[2];
      
      nv1->c[0] += v->c[0];
      nv1->c[1] += v->c[1];
      nv1->c[2] += v->c[2];
      nv1->c[3] += v->c[3];
      
      nv1->n[0] += v->n[0];
      nv1->n[1] += v->n[1];
      nv1->n[2] += v->n[2];
      
    }

    d = 1.0 / (double)(n*n);
    nv1->p[0] *= d;
    nv1->p[1] *= d;
    nv1->p[2] *= d;
    
    nv1->c[0] *= d;
    nv1->c[1] *= d;
    nv1->c[2] *= d;
    nv1->c[3] *= d;
    
    nv1->n[0] *= d;
    nv1->n[1] *= d;
    nv1->n[2] *= d;
    
    // compute the new vertex
    //

    v = vertexPool[vertexID];
    n = vertList.size();
    d = ((double)n - 2.0) / (double)n;

    nv->p[0] = nv0->p[0] + nv1->p[0] + v->p[0]*d;
    nv->p[1] = nv0->p[1] + nv1->p[1] + v->p[1]*d;
    nv->p[2] = nv0->p[2] + nv1->p[2] + v->p[2]*d;
    
    nv->c[0] = nv0->c[0] + nv1->c[0] + v->n[0]*d;
    nv->c[1] = nv0->c[1] + nv1->c[1] + v->n[1]*d;
    nv->c[2] = nv0->c[2] + nv1->c[2] + v->n[2]*d;
    nv->c[3] = nv0->c[3] + nv1->c[3] + v->n[3]*d;
    
    nv->n[0] = nv0->n[0] + nv1->n[0] + v->n[0]*d;
    nv->n[1] = nv0->n[1] + nv1->n[1] + v->n[1]*d;
    nv->n[2] = nv0->n[2] + nv1->n[2] + v->n[2]*d;
    nv->tag = Vertex::Regular;

  }

  dupID = findDuplicateVertex(nv);
  if (dupID < 0) {

    nv->id = vertexPool.size();
    nv->subLevel = subLevel;
    
    newVertexID = nv->id;
    vertexPool.push_back(nv);

  } else {

    newVertexID = dupID;

  }

  delete nv0;
  delete nv1;

  return newVertexID;

}

