/*----------------------------------------------------------
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version
2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied
warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public
License along with this program; if not, write to the Free
Software Foundation, Inc., 59 Temple Place, Suite 330,
Boston, MA  02111-1307  USA

The original author of this program is Hin Jang
March 20, 2003
-----------------------------------------------------------*/

// ----------------------------------------------------------
// quadTreeRoot.cpp
// ----------------------------------------------------------

#include "quadTreeRoot.h"

QuadTreeRoot::QuadTreeRoot()
{

  int i;
  root = new QuadTree();

  for (i=0; i<4; ++i) {

    n[i] = NULL;

  }

  maxSubLevel = 0;
  id = 0;

}

QuadTreeRoot::~QuadTreeRoot()
{

  int i;

  delete root;
  for (i=0; i<4; ++i) {

    if (n[i] != NULL) delete n[i];

  }

}

void QuadTreeRoot::add(int subLevel, Polygon *p)
{

  int i, done;
  QuadTreeRoot *target;
  QuadTree *q, *qn;
  queue<QuadTree*> qt;

  if (subLevel < 0) return;

  target = NULL;
  findForest(this, &target, p->pid);
  if (target != NULL) {

    q = target->root;

    // add polygon to appropriate child
    // of the forest root
    //

    if (subLevel < 2) {

      for (i=0; i<4; ++i) {
	
	if (q->getChild(i) == NULL) {
	  
	  qn = new QuadTree();
	  qn->setSubdivisionLevel(subLevel);
	  qn->setPoly(p);
	  qn->setParent(q);
	  q->setChild(i, qn);
	  break;

	}

      }

    } else {

      done = 0;
      deepAdd(&done, subLevel-1, 0, p, q);

    }

  }

}

void
QuadTreeRoot::deepAdd(int *done, int subLevel, int child,
		      Polygon *p, QuadTree *q)
{

  int i;
  QuadTree *qn, *parent;

  // iterate through quadtree to add polygon
  // to appropriate child
  //

  if (*done || q == NULL) return;
  if (q->getSubdivisionLevel() == subLevel) {

    for (i=0; i<4 && !(*done); ++i) {

      if (q->getChild(i) == NULL) {

	qn = new QuadTree();
	qn->setSubdivisionLevel(subLevel+1);
	qn->setPoly(p);
	qn->setParent(q);
	q->setChild(i, qn);
	*done = 1;

      }
      
    }

    if ( !(*done) ) {

      // traverse to sibling
      //

      if (child < 3) {

	++child;
	parent = q->getParent();
	deepAdd(done, subLevel, child, p,
		parent->getChild(child));
      
      } else {

	// logic flow reaches here for degenerate cases
	//

      }

    }

  } else {

    // traverse to children
    //

    deepAdd(done, subLevel, 0, p, q->getChild(0));
    deepAdd(done, subLevel, 1, p, q->getChild(1));
    deepAdd(done, subLevel, 2, p, q->getChild(2));
    deepAdd(done, subLevel, 3, p, q->getChild(3));

  }

}


void QuadTreeRoot::create()
{

  int polyID;

  polyID = polygonPool[0]->id;
  if (polyID < 1) return;
  createForest(this, 0, polyID);

}

void QuadTreeRoot::createForest(QuadTreeRoot *forest,
				int sourcePolyID,
				int polyID)
{

  int oldSourcePolyID, isLeaf, i;
  QuadTree *qt;
  Polygon *p;

  if (forest == NULL) return;
  isLeaf = 0;

  // tag this forest as a leaf if polyID is negative
  //

  if (polyID < 0) {

    isLeaf = 1;
    polyID = -polyID;

  }

  oldSourcePolyID = sourcePolyID;
  sourcePolyID = polyID;

  // store polygon in quadtree
  //
  
  qt = forest->root;
  p  = polygonPool[polyID-1];
  qt->setPoly(p);
  qt->setSubdivisionLevel(0);
  
  // tag this forest with the id of the polygon
  // belonging to the base mesh
  //
  
  forest->id = p->id;
  
  // terminate forest creation at leaf node
  //

  if (isLeaf) return;

  // populate other forests
  //

  for (i=0; i<4; ++i) {

    if (p->n[i] > 0 &&
	p->n[i] != polyID &&
	p->n[i] != oldSourcePolyID) {
      
      forest->n[i] = new QuadTreeRoot();
      createForest(forest->n[i], sourcePolyID, p->n[i]);
	
    }
    
  }

}

void QuadTreeRoot::findForest(QuadTreeRoot *forest,
			      QuadTreeRoot **target,
			      int polyID)
{

  if (forest == NULL || *target != NULL) return;
  if (forest->id == polyID) {

    *target = forest;

  } else {

    findForest(forest->n[0], target, polyID);
    findForest(forest->n[1], target, polyID);
    findForest(forest->n[2], target, polyID);
    findForest(forest->n[3], target, polyID);

  }

}


void QuadTreeRoot::traverse(int subLevel)
{

  visitForest(subLevel, this);

}


void QuadTreeRoot::visitForest(int subLevel,
			       QuadTreeRoot *forest)
{

  if (forest == NULL) return;
  visitNode(subLevel, forest->root);
  visitForest(subLevel, forest->n[0]);
  visitForest(subLevel, forest->n[1]);
  visitForest(subLevel, forest->n[2]);
  visitForest(subLevel, forest->n[3]);

}

void QuadTreeRoot::visitNode(int subLevel, QuadTree *qt)
{

  QuadTree *t;
  queue<QuadTree*> q;

  if (subLevel < 0 || qt == NULL) return;
  q.push(qt);

  while (q.empty() == false) {

    t = q.front();
    q.pop();

    if (t->getSubdivisionLevel() > subLevel) break;
    if (t->getSubdivisionLevel() == subLevel) {

      // invoke method something like
      // Rasteriser::drawPoly(t->getPoly());
      //

    }

    if (t->getChild(0) != NULL) q.push(t->getChild(0));
    if (t->getChild(1) != NULL) q.push(t->getChild(1));
    if (t->getChild(2) != NULL) q.push(t->getChild(2));
    if (t->getChild(3) != NULL) q.push(t->getChild(3));

  }

}


